/*
 * Copyright (C) 2019 Erlend Sveen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

///////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>		// atoi
#include <string.h>		// memset
#include <sys/socket.h>
#include <netinet/in.h>
#include <spawn.h>		// posix_spawn
#include <sys/stat.h>	// mkfifo
#include <fcntl.h>
#include <unistd.h>		// read write
#include <sys/wait.h>	// waitpid
#include <sys/time.h>	// gettimeofday
#include <errno.h>

///////////////////////////////////////////////////////////////////////////////
// Definitions
///////////////////////////////////////////////////////////////////////////////
#define SOCK_DOMAIN AF_INET

///////////////////////////////////////////////////////////////////////////////
// Extra debug
///////////////////////////////////////////////////////////////////////////////
#define DODEBUG 0

#if DODEBUG
#define DEBUG(str) str
#else
#define DEBUG(str)
#endif

///////////////////////////////////////////////////////////////////////////////
// Data
///////////////////////////////////////////////////////////////////////////////
extern char **environ;
const char *pargs[] = {"sh", 0};

///////////////////////////////////////////////////////////////////////////////
// Functions
///////////////////////////////////////////////////////////////////////////////
int startsh (pid_t *pid, const char *name, char *namestdin, char *namestdout)
{
	// Spawn attributes object with process group
	posix_spawnattr_t attr;
	posix_spawnattr_init (&attr);
	posix_spawnattr_setflags(&attr, POSIX_SPAWN_SETPGROUP);
	posix_spawnattr_setpgroup (&attr, 0);

	// Change input and output file
	posix_spawn_file_actions_t factions;
	posix_spawn_file_actions_init (&factions);
	posix_spawn_file_actions_addopen (&factions, 0, namestdin,
			O_RDONLY, S_IRWXU | S_IRWXG | S_IRWXO);
	posix_spawn_file_actions_addopen (&factions, 1, namestdout,
		O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU | S_IRWXG | S_IRWXO);

	// Launch process!
	int ret = posix_spawn (pid, name, &factions, &attr, (char**)pargs, environ);

	// Delete attr and file actions
	posix_spawnattr_destroy (&attr);
	posix_spawn_file_actions_destroy (&factions);

	// Done
	return ret;
}

void checkforsignals (char *str, int n, int out, pid_t pid)
{
	int wr = 0;

	for (int i = 0; i < n; i++)
	{
		// Disabled because we do not actually have pseudoterminals
		// like /dev/pts
#if 1
		if (str[i] == 0x03)
		{
			short grp = tcgetpgrp (out);
			kill (-grp, SIGINT);
		}
		else if (str[i] == 0x1A)
		{
			short grp = tcgetpgrp (out);
			kill (-grp, SIGTSTP);
		}
		else
#endif
			str[wr++] = str[i];
	}

	DEBUG (printf ("Writing %i %x to pipe\n", wr, str[0]));
	write (out, str, wr);
}

void dosleep (int ms)
{
	struct timespec req;
	req.tv_sec = 0;
	req.tv_nsec = ms * 1000000;
	nanosleep (&req, 0);
}

int process (int sock, int in, int out, pid_t pid)
{
	struct sockaddr_in sa;
	memset ((void*)&sa, 0, sizeof (struct sockaddr_in));
	socklen_t len = sizeof (struct sockaddr_in);
	int hasclient = 0;
	char linebuf[128];
	unsigned int lastsendtime = 0;

	while (1)
	{
		// Socket to stdin fifo
		int rd = recvfrom (sock, linebuf, sizeof(linebuf), 0,
			(struct sockaddr*) &sa, &len);

		if (rd > 0)
		{
			DEBUG (printf ("Got %i data\n", rd));
			checkforsignals (linebuf, rd, in, pid);
			//write (in, linebuf, rd);
			//write (in, "\r", 1);
			hasclient = 1;
		}

		// stdout fifo to socket
		int wr = read (out, linebuf, sizeof (linebuf)/2);

		if (wr > 0 && hasclient)
		{
			DEBUG (printf ("Sending data to client: "));
			DEBUG (for (int i = 0; i < wr; i++) printf ("0x%x ", linebuf[i]));
			DEBUG (printf ("\n"));

			for (int i = 0; i < 500/2; i++)
			{
				struct timeval now;
				gettimeofday (&now, 0);
				unsigned int intnow = (now.tv_sec * 1000) + (now.tv_usec / 1000);
				unsigned int diff = intnow - lastsendtime;
				lastsendtime = intnow;

				if (diff <= 5)
					dosleep (5);

				if (sendto (sock, linebuf, wr, 0,
						(struct sockaddr*) &sa, len) >= 0)
					break;
				else
					dosleep (2);
			}
		}

		// Check if terminated
		int status = 0;

		if (waitpid (pid, &status, WNOHANG) > 0)
			if (WIFEXITED(status))
				return 6;

		// Sleep if nothing happened
		if (rd <= 0 && wr <= 0)
		{
			dosleep (3);
			continue;
		}
	}
}

int realmain (int argc, char **argv, char *namea, char *nameb)
{
	// Start the shell first, to avoid duplicating file descriptors
	pid_t p;
	char *name = argc >= 4 ? argv[3] : "/bin/sh";

	if (startsh (&p, name, namea, nameb) < 0)
		return 4;

	// Open FIFO for stdin and stdout
	int a = mkfifo (namea, 0x1A4);
	int b = mkfifo (nameb, 0x1A4);

	if (a < 0)
		return 3;

	if (b < 0)
		return 33;

	// Socket and socket address
	struct sockaddr_in sa;
	memset ((void*)&sa, 0, sizeof (struct sockaddr_in));
	sa.sin_family = SOCK_DOMAIN;
	sa.sin_addr.s_addr = htonl (INADDR_ANY);
	sa.sin_port = htons (22);

	if (argc >= 2)
	{
		printf ("Using port %i\n", atoi (argv[1]));
		//fputs ("Using port %i\n", stdout);
		//fputs (argv[1], stdout);
		//fputs ("\n", stdout);
		sa.sin_port = htons (atoi (argv[1]));
	}

	int s = socket (SOCK_DOMAIN, SOCK_DGRAM, IPPROTO_UDP);

	if (s < 0)
		return 1;

	// Set non-blocking operation
	int f = fcntl (a, F_GETFL);
	fcntl (a, F_SETFL, f | O_NONBLOCK);
	f = fcntl (b, F_GETFL);
	fcntl (b, F_SETFL, f | O_NONBLOCK);

	f = fcntl (s, F_GETFL);
	fcntl (s, F_SETFL, f | O_NONBLOCK);

	// Set process group of both pipes (non-standard)
	tcsetpgrp (a, p);
	tcsetpgrp (b, p);

	// Bind the socket
	if (bind (s, (struct sockaddr*) &sa, sizeof (struct sockaddr_in)) < 0)
		return 2;

	// Cheap way to set hostname of session
	if (argc >= 3)
	{
		const char hn[] = "sethn ";
		write (a, hn, strlen (hn));
		write (a, argv[2], strlen (argv[2]));
		write (a, "\r", 1);
	}
	else
	{
		const char hn[] = "sethn nssh\r";
		write (a, hn, strlen (hn));
	}

	// Process data between pipes and socket
	int r = process (s, a, b, p);
	close (a);
	close (b);
	return r;
}

int main (int argc, char **argv)
{
	// Create FIFO names
	char namea[32];
	char nameb[32];
	sprintf (namea, "/pipe/nsshs-stdin-%i", (int) getpid());
	sprintf (nameb, "/pipe/nsshs-stdout-%i", (int) getpid());

#if 1
	int r = realmain (argc, argv, namea, nameb);
	printf ("Main returned %i, errno %i\n", r, errno);
#else
	int r = realmain (argc, argv, namea, nameb);
#endif

	unlink (namea);
	unlink (nameb);

	return r;
}

/* vim: noexpandtab set ts=4 sw=4 sts=4 tw=80 :*/
